#
# Extensible User Folder
#
# (C) Copyright 2000,2001 The Internet (Aust) Pty Ltd
# ACN: 082 081 472  ABN: 83 082 081 472
# All Rights Reserved
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Author: Andrew Milton <akm@theinternet.com.au>

from time import time
from Sync import Synchronized
from BTrees.OOBTree import OOBTree

class UserCacheItem:
	lastAccessed=0
	def __init__(self, username, password, User):
		self.username=username
		self.password=password
		self.userObj = User
		self.lastAccessed=time()

	def touch(self):
		self.lastAccessed=time()
		
	def __repr__(self):
		return self.username

SessionExpiredException='User Session Expired'
class UserCache:
	def __init__(self, sessionLength):
		self.sessionLength=sessionLength
		self.cache=OOBTree()
		self.hits=0
		self.fail=0
		self.nouser=0
		self.attempts=0
		self.timeouts=0
		self.cacheStarted=time()

	def addToCache(self, username, password, User):
		if not self.sessionLength:
			return

		try:
			u = self.cache.items(username)
			if u:
				for x in self.cache[username]:
					self.cache.remove(x)
		except:
			pass
			
		u = UserCacheItem(username, password, User)
		self.cache[username]=u

	def getUser(self, username, password):
		
		if not self.sessionLength:
			return None

		self.attempts=self.attempts+1
		
		u = None
		try:
			u = self.cache[username]
		except KeyError:
			self.nouser=self.nouser+1
			return None

		now = time()
		if u:
			if u.password != password:
				self.fail=self.fail+1
				del self.cache[u.username]
			elif self.sessionLength and (
				(now - u.lastAccessed) > self.sessionLength):
				del self.cache[u.username]
				self.timeouts=self.timeouts+1
				u.userObj.notifyCacheRemoval()
				del u.userObj
				del u
				raise SessionExpiredException
			else:
				u.touch()
				self.hits=self.hits+1
				return u.userObj

		self.nouser=self.nouser+1
		return None

	def removeUser(self, username):
		if not self.sessionLength:
			return
		try:
			if self.cache[username]:
				del self.cache[username]
		except:
			pass

	def getCacheStats(self):
		return (
			{'attempts':self.attempts,
			 'hits':self.hits,
			 'fail':self.fail,
			 'misses':self.nouser,
			 'cachesize':len(self.cache),
			 'time':self.cacheStarted,
			 'timeouts':self.timeouts,
			 'length':self.sessionLength})
		
	def getCurrentUsers(self):
		x=[]
		now = time()		
		for z in self.cache.keys():
			u = self.cache[z]
			if self.sessionLength and (
				(now - u.lastAccessed) > self.sessionLength):
				del self.cache[u.username]
				self.timeouts=self.timeouts+1
				u.userObj.notifyCacheRemoval()
				del u.userObj
				del u
			else:
				x.append({'username':u.username,
						  'lastAccessed':u.lastAccessed})
		return x
