#
# Extensible User Folder
#
# (C) Copyright 2000,2001 The Internet (Aust) Pty Ltd
# ACN: 082 081 472  ABN: 83 082 081 472
# All Rights Reserved
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Author: Andrew Milton <akm@theinternet.com.au>

##############################################################################
#
# Zope Public License (ZPL) Version 0.9.4
# ---------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
# 
# 1. Redistributions in source code must retain the above
#    copyright notice, this list of conditions, and the following
#    disclaimer.
# 
# 6. Redistributions of any form whatsoever must retain the
#    following acknowledgment:
# 
#      "This product includes software developed by Digital
#      Creations for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND
#   ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
#   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT
#   SHALL DIGITAL CREATIONS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
#   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
#   IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
#   THE POSSIBILITY OF SUCH DAMAGE.
#
##############################################################################

import Globals, App.Undo, socket, os, string

from Globals import DTMLFile, HTMLFile, MessageDialog, INSTANCE_HOME
from string import join,strip,split,lower,upper,find
from App.Management import Navigation, Tabs
from Acquisition import Implicit
from OFS.SimpleItem import Item
from OFS.Document import Document
from OFS.Folder import Folder
from OFS.CopySupport import CopyContainer
from base64 import decodestring, encodestring
from urllib import quote, unquote
from ImageFile import ImageFile
from AccessControl.Role import RoleManager
from AccessControl.User import BasicUser, BasicUserFolder, readUserAccessFile
from OFS.DTMLMethod import DTMLMethod
from Products.ZSQLMethods.SQL import SQLConnectionIDs
from time import time

from PropertyEditor import *

from UserCache.UserCache import UserCache, SessionExpiredException
from LoginRequiredMessages import LoginRequiredMessages

class User(BasicUser):

	icon='misc_/exUserFolder/exUser.gif'

	def __init__(self, name, password, roles, domains, propSource,
				 cryptPassword):
		self.name   =name
		self.__     =password
		self.roles  =filter(None, roles)
		self.domains=filter(None, map(strip, split(domains, ',')))
		self.propSource=propSource
		self.cryptPassword=cryptPassword

	def getUserName(self):
		return self.name

	def _getPassword(self):
		return self.__

	def getRoles(self):
		return tuple(self.roles) + ('Authenticated',)
    
	def getDomains(self):
		return self.domains

	# Ultra generic way of getting, checking and setting properties
	def getProperty(self, property, default=None):
		if self.propSource:
			return self.propSource.getProperty(property, default)

	def hasProperty(self, property):
		if self.propSource:
			return self.propSource.hasProperty(property)

	def setProperty(self, property, value):
		if property[0]=='_':
			return
		if self.propSource:
			return self.propSource.setProperty(property, value)

	def setTempProperty(self, property, value):
		if property[0]=='_':
			return
		if self.propSource:
			return self.propSource.setTempProperty(property, value)

	def flushTempProperties(self):
		if self.propSource:
			return self.propSource.flushTempProperties()

	def delProperty(self, property):
		if property[0]=='_':
			return
		if self.propSource:
			return self.propSource.delProperty(property, value)
		
	def listProperties(self):
		if self.propSource:
			return self.propSource.listProperties()

	# Try to allow User['property'] -- won't work for password d;)
	def __getitem__(self, key):
		# Don't return 'private' keys
		if key[0] != '_':
			if hasattr(self, key):
				return getattr(self, key)
			if self.propSource and self.propSource.hasProperty(key):

				return self.propSource.getProperty(key)
		raise KeyError, key

	def __setitem__(self, key, value):
		if key[0]=='_':
			return
		if self.propSource:
			self.propSource.setProperty(key, value)
		
	# List one user is supplied by the Auth Source...
	
	def authenticate(self, listOneUser, password, request, remoteAuth=None):
		result=listOneUser(username=self.name)
		for people in result:
			if remoteAuth:
				return remoteAuth(self.name, password)
			else:
				secret=self.cryptPassword(self.name, password)
				return secret==people['password']
		return None

	# You can set logout times or whatever here if you want to, the
	# property source is still active.
	def notifyCacheRemoval(self):
		if self.propSource:
			self.propSource.flushTempProperties()

# Little function to create temp usernames
def createTempName():
	t=time()
	t1=time()
	t2=time()
	t3 = 0.0
	t3 = (t + t1 + t2) / 3
	un = "Anonymous %.0f"%(t3)
	return(un)
#
# An Anonymous User for session tracking...
# Can set and get properties just like a normal user.
#
# These objects live in the cache, so, we have a __del__ method to
# clean ourselves up.
#

class AnonUser(User):
	def __init__(self, name, roles, propSource):
		self.name   =name
		self.__     =''
		self.roles  =filter(None, roles)
		self.propSource=propSource

	def getRoles(self):
		return tuple(self.roles) + ('Anonymous',)

	def authenticate(self, listOneUser, password, request, remoteAuth=None):
		return 1
	
	def notifyCacheRemoval(self):
		if self.propSource:
			self.propSource.deleteUsers([self.name,])

manage_addexUserFolderForm=DTMLFile('dtml/manage_addexUserFolder', globals())

def manage_addexUserFolder(self, authId, propId, memberId, 
						   cookie_mode=0, session_length=0,
						   sessionTracking=None, idleTimeout=None,
						   REQUEST={}):
	""" """
	if hasattr(self.aq_base, 'acl_users'):
		return MessageDialog(
			title  ='Item Exists',
			message='This object already contains a User Folder',
			action ='%s/manage_main' % REQUEST['URL1'])
	ob=exUserFolder(authId, propId, memberId, cookie_mode, session_length,
					sessionTracking, idleTimeout)
	self._setObject('acl_users', ob, None, None, 0)
	self.__allow_groups__=self.acl_users
	ob=getattr(self, 'acl_users')
	ob.postInitialisation(REQUEST)
	if REQUEST:
		return self.manage_main(self, REQUEST)
	return ''

class exUserFolder(Folder,BasicUserFolder,CopyContainer):
	""" """

	# HACK! We use this meta_type internally so we can be pasted into
	# the root. We registered with 'exUserFolder' meta_type however, so
	# our constructors work.
	meta_type='User Folder'
	id       ='acl_users'
	title    ='Extensible User Folder'
	icon     ='misc_/exUserFolder/exUserFolder.gif'

	isPrincipiaFolderish=1
	isAUserFolder=1

	authSources={}
	propSources={}
	membershipSources={}

	manage_options=(
		{'label':'Users',      'action':'manage_main'},
		{'label':'Parameters', 'action':'manage_editexUserFolderForm'},
		{'label':'Authentication Source','action':'manage_editAuthSourceForm'},
		{'label':'Properties Source','action':'manage_editPropSourceForm'},
		{'label':'Membership Source', 'action':'manage_editMembershipSourceForm'},
		{'label':'Cache Data', 'action':'manage_showCacheData'},
		{'label':'Security',   'action':'manage_access'},
		{'label':'Contents',   'action':'manage_contents'},
		{'label':'Ownership',  'action':'manage_owner'},
		{'label':'Undo',       'action':'manage_UndoForm'},
		)

	__ac_permissions__=(
		('View management screens',
		 ('manage','manage_menu','manage_main','manage_copyright',
		  'manage_tabs', 'manage_properties', 'manage_UndoForm',
		  'manage_edit', 'manage_contents','manage_cutObjects',
		  'manage_copyObjects', 'manage_pasteObjects',
          'manage_renameForm', 'manage_renameObject',
		  'manage_renameObjects', )),
		
		('Undo changes',       ('manage_undo_transactions')),
		('Change permissions', ('manage_access')),
		('Manage users',       ('manage_users', 'manage_editUserForm', 'manage_editUser', 'manage_addUserForm', 'manage_addUser', 'manage_userActions')),
		('Change ex User Folders',     ('manage_edit')),
		('View', ('manage_changePassword', 'manage_forgotPassword', 'docLogin', 'docLogout', 'logout', 'DialogHeader', 'DialogFooter', 'manage_signupUser')),
		('Manage properties', ('manage_addProperty',
							   'manage_editProperties',
							   'manage_delProperties',
							   'manage_changeProperties',
							   'manage_propertiesForm',
							   'manage_propertyTypeForm',
							   'manage_changePropertyTypes',
							   )),
		('Access contents information',
		 ('hasProperty', 'propertyIds', 'propertyValues','propertyItems',
		  'getProperty', 'getPropertyType', 'propertyMap', ''),
		 ('Anonymous', 'Manager'),
		 ),
		)
	manage_access=DTMLFile('dtml/access',globals())
	my_manage_tabs=DTMLFile('common/manage_tabs',globals())
	manage_properties=DTMLFile('dtml/properties', globals())
	manage_main=DTMLFile('dtml/mainUser', globals())
	manage_contents=Folder.manage_main
	MessageDialog=DTMLFile('common/MessageDialog', globals())
	manage_showCacheData=DTMLFile('dtml/manage_showCacheData', globals())
	docLoginRedirect=DTMLFile('dtml/docLoginRedirect', globals())
	docLoginRedirect__roles__=()

	# Stupid crap
	try:
		manage_contents._setName('manage_contents')
	except AttributeError:
		pass

	

	manage_addUserForm=DTMLFile('dtml/manage_addUserForm',globals())
	manage_editUserForm=DTMLFile('dtml/manage_editUserForm',globals())

	DialogHeader=DTMLFile('common/DialogHeader',globals())
	DialogFooter=DTMLFile('common/DialogFooter',globals())

	manage_editAuthSourceForm=DTMLFile('dtml/manage_editAuthSourceForm',globals())
	manage_editPropSourceForm=DTMLFile('dtml/manage_editPropSourceForm',globals())
	manage_editMembershipSourceForm=DTMLFile('dtml/manage_editMembershipSourceForm', globals())

	manage_addPropertyForm=DTMLFile('dtml/manage_addPropertyForm', globals())
	manage_createPropertyForm=DTMLFile('dtml/manage_createPropertyForm', globals())
	manage_editUserPropertyForm=DTMLFile('dtml/manage_editUserPropertyForm', globals())

	manage_editexUserFolderForm=DTMLFile('dtml/manage_editexUserFolderForm', globals())

	# Yeee hah I should really take my own advice and use __setstate_
	# But that's a tad expensive for a user folder.
	_v_UserCache=None
	sessionLength=0 # Upgrading users should get no caching.
	cookie_mode=0
	sessionTracking=None # Or session tracking.
	idleTimeout=None

	def __init__(self, authId, propId, memberId, cookie_mode=0,
				 session_length=0, sessionTracking=None, idleTimeout=None):
		self.cookie_mode=cookie_mode
		self.sessionLength=session_length
		self.sessionTracking=sessionTracking
		self.idleTimeout=idleTimeout
		
		_docLogin=DTMLFile('dtml/docLogin',globals())
		_docLogout=DTMLFile('dtml/docLogout',globals())

		docLogin=DTMLMethod(__name__='docLogin')
		docLogin.manage_edit(data=_docLogin, title='Login Page')
		self._setObject('docLogin', docLogin, None, None, 0)

		docLogout=DTMLMethod(__name__='docLogout')
		docLogout.manage_edit(data=_docLogout, title='Logout Page')
		self._setObject('docLogout', docLogout, None, None, 0)

		postUserCreate=DTMLMethod(__name__='postUserCreate')
		postUserCreate.manage_edit(data=_postUserCreate, title='Post User Creation methods')
		self._setObject('postUserCreate', postUserCreate, None, None, 0)

		self.manage_addAuthSource=self.authSources[authId].manage_addMethod
		self.manage_addPropSource=self.propSources[propId].manage_addMethod
		self.manage_addMembershipSource=self.membershipSources[memberId].manage_addMethod

	def manage_beforeDelete(self, item, container):
		if item is self:
			try:
				del container.__allow_groups__
			except:
				pass

	def manage_editPropSource(self, REQUEST):
		""" Edit Prop Source """
		if self.currentPropSource:
			self.currentPropSource.manage_editPropSource(REQUEST)
		return self.manage_main(self, REQUEST)

	def manage_editAuthSource(self, REQUEST):
		""" Edit Auth Source """
		self.currentAuthSource.manage_editAuthSource(REQUEST)
		return self.manage_main(self, REQUEST)

	def manage_editMembershipSource(self, REQUEST):
		""" Edit Membership Source """
		if self.currentMembershipSource:
			return self.currentMembershipSource.manage_editMembershipSource(REQUEST)
	
	def postInitialisation(self, REQUEST):
		self.manage_addAuthSource(self=self,REQUEST=REQUEST)
		self.manage_addPropSource(self=self,REQUEST=REQUEST)
		self.manage_addMembershipSource(self=self,REQUEST=REQUEST)
	
	def addAuthSource(self, REQUEST={}):
		return self.manage_addAuthSourceForm(self, REQUEST)

	def addPropSource(self, REQUEST={}):
		return self.manage_addPropSourceForm(self, REQUEST)

	def addMembershipSource(self, REQUEST={}):
		return self.manage_editMembershipSourceForm(self, REQUEST)

	def listUserProperties(self, username):
		if self.currentPropSource:
			return self.currentPropSource.listUserProperties(username=username)

	def getUserProperty(self, username, key):
		if self.currentPropSource:
			return self.currentPropSource.getUserProperty(key=key, username=username)
	
	def reqattr(self, request, attr):
		try:    return request[attr]
		except: return None

	def getAuthFailedMessage(self, code):
		""" Return a code """
		if LoginRequiredMessages.has_key(code):
			return LoginRequiredMessages[code]
		return 'Login Required'

	def manage_editUser(self, username, REQUEST={}):
		""" Edit a User """
		# username=self.reqattr(REQUEST,'username')
		password=self.reqattr(REQUEST,'password')
		password_confirm=self.reqattr(REQUEST,'password_confirm')
		roles=self.reqattr(REQUEST,'roles')

		if not username:
			return MessageDialog(
				title  ='Illegal value', 
				message='A username must be specified',
				action ='manage_main')

		if (password or password_confirm) and (password != password_confirm):
			return MessageDialog(
				title  ='Illegal value', 
				message='Password and confirmation do not match',
				action ='manage_main')

		self.currentAuthSource.updateUser(username, password, roles)

		if self.currentPropSource:
			self.currentPropSource.updateUser(username, REQUEST)

		# We may have updated roles or passwords... flush the user...
		if self._v_UserCache:
			# If the cache isn't here, there's no point creating it just to
			# remove something that won't be in it.
			self._v_UserCache.removeUser(username)			
		
		return MessageDialog(
			title = 'User Updated',
			message= 'User %s was updated.'%(username),
			action = 'manage_main')


	#
	# Membership helper
	#
	def goHome(self, REQUEST, RESPONSE):
		""" Go to home directory """
		if self.currentMembershipSource:
			self.currentMembershipSource.goHome(REQUEST, RESPONSE)
	#
	# Membership method of adding a new user.
	# If everything goes well the membership plugin calls manage_addUser()
	#
	def manage_signupUser(self, REQUEST):
		""" Signup a new user """
		""" This is seperate so you can add users using the normal """
		""" interface w/o going through membership policy """

		username=self.reqattr(REQUEST,'username')
		roles=self.reqattr(REQUEST,'roles')

		if not username:
			return MessageDialog(
				title  ='Illegal value', 
				message='A username must be specified',
				action ='manage_main')

		if (self.getUser(username) or
			(self._emergency_user and
			 username == self._emergency_user.getUserName())):
			return MessageDialog(
				title  ='Illegal value', 
				message='A user with the specified name already exists',
				action ='manage_main')

		if self.currentMembershipSource:
			return self.currentMembershipSource.createUser(REQUEST)

	#
	# Membership method of changing passwords
	#
	def manage_changePassword(self, REQUEST):
		""" Change a password """
		if self.currentMembershipSource:
			return self.currentMembershipSource.changePassword(REQUEST)
		
	#
	# User says they can't remember their password
	#
	def manage_forgotPassword(self, REQUEST):
		""" So something about forgetting your password """
		if self.currentMembershipSource:
			return self.currentMembershipSource.forgotPassword(REQUEST)
		
	def __creatable_by_emergency_user__(self): return 1

	def manage_addUser(self, REQUEST):
		""" Add a New User """
		username=self.reqattr(REQUEST,'username')
		password=self.reqattr(REQUEST,'password')
		password_confirm=self.reqattr(REQUEST,'password_confirm')
		roles=self.reqattr(REQUEST,'roles')

		if not username:
			return MessageDialog(
				title  ='Illegal value', 
				message='A username must be specified',
				action ='manage_main')

		if not password or not password_confirm:
			return MessageDialog(
				title  ='Illegal value', 
				message='Password and confirmation must be specified',
				action ='manage_main')

		if (self.getUser(username) or
			(self._emergency_user and
			 username == self._emergency_user.getUserName())):
			return MessageDialog(
				title  ='Illegal value', 
				message='A user with the specified name already exists',
				action ='manage_main')

		if (password or password_confirm) and (password != password_confirm):
			return MessageDialog(
				title  ='Illegal value', 
				message='Password and confirmation do not match',
				action ='manage_main')

		self.currentAuthSource.createUser(username, password, roles)
			
		# REQUEST holds all the vars from the form
		# That way we can use a generic entry point.
		if self.currentPropSource:
			self.currentPropSource.createUser(username, REQUEST)

		#
		# Explicitly check our contents, do not just acquire postUserCreate
		#
		if 'postUserCreate' in self.objectIds():
			self.postUserCreate(self, REQUEST)
		
		return MessageDialog(
			title = 'User Created',
			message= 'User %s was created.'%(username),
			action = 'manage_main')

	def _doAddUser(self, name, password, roles, domains=''):
		""" For programatically adding simple users """
		self.currentAuthSource.createUser(name, password, roles)
		if self.currentPropSource:
			self.currentPropSource.createUser(name, {})

	def _doChangeUser(self, name, password, roles, domains=''):
		self.currentAuthSource.updateUser(username, password, roles)

	def _doDelUsers(self, names):
		self.deleteUsers(names)

	def _createInitialUser(self):
		if len(self.getUserNames()) <= 1:
			info = readUserAccessFile('inituser')
			if info:
				name, password, domains, remote_user_mode = info
				self._doAddUser(name, password, ('Manager',), domains)
	
	def getUsers(self):
		"""Return a list of user objects or [] if no users exist"""
		data=[]
		try:
			items=self.listUsers()
			for people in items:
				user=User(people['username'],
						  people['password'],
						  people['roles'], '',
						  self.currentPropSource, self.cryptPassword)
				data.append(user)
		except:
			pass
			
		return data

	def getUser(self, name):
		"""Return the named user object or None if no such user exists"""
		try:
			items=self.listOneUser(name)
		except:
			import traceback
			traceback.print_exc()
			return None

		if not items:
			return None
		
		for people in items:
			return User(people['username'], people['password'], people['roles'], '', self.currentPropSource, self.cryptPassword)
		return None
		
	def manage_userActions(self, submit=None, userids=None, REQUEST={}):
		""" Do things to users """
		if submit==' Add... ':
			return self.manage_addUserForm(self, REQUEST)
		if submit==' Delete ':
			self.deleteUsers(userids)
			return MessageDialog(
				title  ='Users Deleted',
				message='Selected Users have been deleted',
				action =REQUEST['URL1']+'/manage_main',
				target ='manage_main')

		if REQUEST:
			return self.manage_main(self,REQUEST)
		return ''
	

	def std_validate(self,request,auth='',roles=None):
		parent=request['PARENTS'][0]

		# If no authorization, only a user with a
		# domain spec and no passwd or nobody can
		# match
		if not auth:
			for ob in self.getUsers():
				domains=ob.getDomains()
				if domains:
					if ob.authenticate(self.currentAuthSource.listOneUser, '',
									   request,
									   self.currentAuthSource.remoteAuthMethod):
						if ob.allowed(parent, roles):
							ob=ob.__of__(self)
							return ob

			nobody=self._nobody
			if self._isTop() and nobody.allowed(parent, roles):
				ob=nobody.__of__(self)
				return ob
			return None

		# Only do basic authentication
		if lower(auth[:6])!='basic ':
			return None
		name,password=tuple(split(decodestring(split(auth)[-1]), ':'))

		# Check for emergency user
		emergency=self._emergency_user
		if emergency and name==emergency.getUserName():
			if self._isTop():
				return emergency.__of__(self)
			else:
				return None # Not top level folder.
			
		#
		# User has logged in recently
		# ignore authentication...
		#
		try:
			user = self._v_UserCache.getUser(name, password)
			if user:
				return user
		except SessionExpiredException:
			if self.idleTimeout:
				return None

		# Try to get user
		user=self.getUser(name)
		if user is None:
			return None

		# Try to authenticate user
		if not user.authenticate(self.currentAuthSource.listOneUser,
								 password, request,
								 self.currentAuthSource.remoteAuthMethod):
			return None

		# We need the user to be able to acquire!
		user=user.__of__(self)

		# Try to authorize user
		if user.allowed(parent, roles):
			self._v_UserCache.addToCache(name, password, user)
			return user
		return None
		
	def redirectToLogin(self, REQUEST):
		""" Allow methods to call from Web """
		script=''
		pathinfo=''
		querystring=''
		redirectstring=''

		if not REQUEST.has_key('destination'):
			if self.currentMembershipSource:
				redirectstring = self.currentMembershipSource.getLoginDestination(REQUEST)
			else:
				script=REQUEST['SCRIPT_NAME']
				pathinfo=REQUEST['PATH_INFO']
				if REQUEST.has_key('QUERY_STRING'):
					querystring='?'+REQUEST['QUERY_STRING']

				redirectstring=script+pathinfo
				if querystring:
					redirectstring=redirectstring+querystring

			REQUEST['destination']=redirectstring
		if self.currentMembershipSource and self.currentMembershipSource.loginPage:
			try:
				REQUEST.RESPONSE.redirect('%s/%s%s'%(
					self.currentMembershipSource.baseURL,
					self.currentMembershipSource.loginPage,
					redirectstring))
				return
			except:
				pass
		return self.docLogin(self,REQUEST)

	def cookie_validate(self,request,auth='',roles=None):
		parent=request['PARENTS'][0]
		req_has=request.has_key
		resp=request['RESPONSE']

		if req_has('__ac_name') and req_has('__ac_password'):
			name=request['__ac_name']
			password=request['__ac_password']

			#
			# User has logged in recently
			# ignore authentication...
			#
			try:
				user = self._v_UserCache.getUser(name, password)
				if user:
					return user
			except SessionExpiredException:
				if self.idleTimeout:
					request.set('authFailedCode', 'session_expired')
					raise 'LoginRequired', self.docLoginRedirect(self, request)
			
			# Find user
			user=self.getUser(name)
			if user is not None:
				if user.authenticate(self.currentAuthSource.listOneUser,
									 password, request,
									 self.currentAuthSource.remoteAuthMethod):
					user=user.__of__(self)

					if user.allowed(parent, roles):
						token='%s:%s' % (name, password)
						token=encodestring(token)
						token=quote(token)
						resp.setCookie('__ac', token, path='/')
						request['__ac']=token
						try:
							del request['__ac_name']
							del request['__ac_password']
						except:
							pass
						self._v_UserCache.addToCache(name, password, user)
						return user
					else:
						request.set('authFailedCode', 'unauthorized')
				else:
					request.set('authFailedCode', 'login_failed')
			else:
				request.set('authFailedCode','login_failed')

			raise 'LoginRequired', self.docLoginRedirect(self, request)

		if req_has('__ac'):
			c=request['__ac']
			c=unquote(c)
			try:
				c=decodestring(c)
			except:
				resp.expireCookie('__ac', path='/')
				raise 'LoginRequired', self.docLoginRedirect(self, request)
            
			name,password=tuple(split(c, ':'))

			# Check for emergency user
			emergency=self._emergency_user
			if emergency and name==emergency.getUserName():
				if self._isTop():
					return emergency.__of__(self)
				else:
					return None # Not top level folder.

			#
			# User has logged in recently
			# ignore authentication...
			#
			try:
				user = self._v_UserCache.getUser(name, password)			
				if user:
					return user
				#
				# If we get to here and we have the cookie set,
				# and we weren't in the cache, the cache must have been cleaned
				# so we consider that a session timeout.
				#
				if self.idleTimeout:
					#
					# Remove the cookie to prevent looping
					#
					resp.expireCookie('__ac', path='/')
					request.set('authFailedCode', 'session_expired')
					raise 'LoginRequired', self.docLoginRedirect(self, request)
			except SessionExpiredException:
				if self.idleTimeout:
					#
					# Remove the cookie to prevent looping
					#
					resp.expireCookie('__ac', path='/')
					request.set('authFailedCode', 'session_expired')
					raise 'LoginRequired', self.docLoginRedirect(self, request)

			# Find user
			user=self.getUser(name)
			if user is not None:
				if user.authenticate(self.currentAuthSource.listOneUser,
									 password, request,
									 self.currentAuthSource.remoteAuthMethod):
					user=user.__of__(self)
					if user.allowed(parent, roles):
						self._v_UserCache.addToCache(name, password, user)
						return user
					else:
						request.set('authFailedCode', 'unauthorized')
				else:
					request.set('authFailedCode', 'login_failed')

			elif self.sessionTracking and self.currentPropSource:
				# We got to here because a cookie was set, but,
				# we couldn't find the user... a timed out anonymous
				# user.
				ob = self.createAnonymousUser(request, resp)
				return ob
			
			resp.expireCookie('__ac', path='/')
			raise 'LoginRequired', self.docLoginRedirect(self, request)

        
		if auth: return None

		# We don't cache from here down...

		# Try domain matching
		for ob in self.getUsers():
			domains=ob.getDomains()
			if ob.domains:
				if ob.authenticate(self.currentAuthSource.listOneUser, '',
								   request,
								   self.currentAuthSource.remoteAuthMethod):
					if ob.allowed(parent, roles):
						ob=ob.__of__(self)
						return ob

		if not self.sessionTracking:
			# Allow anonymous access to things under exUserFolders,
			# if the object is public.
			nobody=self._nobody
			if nobody.allowed(parent, roles):
				ob=nobody.__of__(self)
				return ob
		elif self.currentPropSource:
			ob = self.createAnonymousUser(request, resp)
			return ob

		raise 'LoginRequired', self.docLoginRedirect(self, request)

	def setAnonCookie(self, name, request, resp):
		token='%s:%s' % (name, '')
		token=encodestring(token)
		token=quote(token)
		resp.setCookie('__ac', token, path='/')
		request['__ac']=token

	def createAnonymousUser(self, request, resp):
		aName=createTempName()
		bogusREQUEST={}
		bogusREQUEST['user_realname']='Guest User'
		self.currentPropSource.createUser(aName, bogusREQUEST)
		ob = AnonUser(aName, [], self.currentPropSource)
		ob = ob.__of__(self)
		self._v_UserCache.addToCache(aName, '', ob)			
		self.setAnonCookie(aName, request, resp)
		return ob
		
	def validate(self,request,auth='',roles=None):
		if self._v_UserCache==None:
			self._v_UserCache=UserCache(self.sessionLength)

		# If we get 'auth' and cookie_mode it's probably an FTP connect
		# or we're authenticating from a higher user folder that doesn't
		# use cookies.. since we're validating anyway, I don't see this
		# as a security hole.
		if self.cookie_mode and not auth:
			return self.cookie_validate(request, auth, roles)
		else:
			return self.std_validate(request, auth, roles)

	def __len__(self):
		return len(self.listUsers())


	def manage_edit(self, cookie_mode, session_length, sessionTracking=None,
					idleTimeout=None, REQUEST=None):
		"""Change properties"""

		self.cookie_mode=cookie_mode
		self.sessionLength=session_length
		self.sessionTracking=sessionTracking
		self.idleTimeout=idleTimeout
		
		if REQUEST:
			return self.MessageDialog(
				title  ='exUserFolder Changed',
				message='exUserFolder properties have been updated',
				action =REQUEST['URL1']+'/manage_main',
				target ='manage_main')

	def logout(self, REQUEST):
		"""Logout"""
		if self._v_UserCache:
			self._v_UserCache.removeUser(REQUEST['AUTHENTICATED_USER'].getUserName())
		REQUEST['RESPONSE'].expireCookie('__ac', path='/')
		return self.docLogout(self, REQUEST)

	#
	# Methods to be supplied by Auth Source
	#
	def deleteUsers(self, userids):
		self.currentAuthSource.deleteUsers(userids)

		# Comment out to use Andreas' pgSchema
		if self.currentPropSource:
			self.currentPropSource.deleteUsers(userids)

	def listUsers(self):
		return self.currentAuthSource.listUsers()

	def user_names(self):
		return self.currentAuthSource.listUserNames()
	
	def getUserNames(self):
		return self.currentAuthSource.listUserNames()

	def listOneUser(self,username):
		return self.currentAuthSource.listOneUser(username)

	def cryptPassword(self, username, password):
		return self.currentAuthSource.cryptPassword(username, password)

	def PropertyEditor(self):
		""" """
		if self.REQUEST.has_key(self.REQUEST['propName']):
			return PropertyEditor.EditMethods[self.REQUEST['propType']](self.REQUEST['propName'], self.REQUEST[self.REQUEST['propName']])
		return PropertyEditor.EditMethods[self.REQUEST['propType']](self.REQUEST['propName'], None)

	def PropertyView(self):
		""" """
		if self.REQUEST.has_key(self.REQUEST['propName']):
			return PropertyEditor.ViewMethods[self.REQUEST['propType']](self.REQUEST['propName'], self.REQUEST[self.REQUEST['propName']])
		return PropertyEditor.ViewMethods[self.REQUEST['propType']](self.REQUEST['propName'], None)

	def manage_addUserProperty(self, username, propName, propValue, REQUEST):
		""" add a new propery """

		self.currentPropSource.setUserProperty(propName, username, propValue)
		return self.manage_editUserForm(self,REQUEST)

	def getUserCacheStats(self):
		""" Stats """
		if self.sessionLength and self._v_UserCache:
			if self._v_UserCache.getCacheStats()['attempts']:
				return self._v_UserCache.getCacheStats()
		return None

	def getUserCacheUsers(self):
		""" Current Users """
		if self.sessionLength and self._v_UserCache:
			return self._v_UserCache.getCurrentUsers()
		return None

def doAuthSourceForm(self,authId):
	""" la de da """
	return exUserFolder.authSources[authId].manage_addForm

def doPropSourceForm(self,propId):
	""" la de da """
	return exUserFolder.propSources[propId].manage_addForm

def doMembershipSourceForm(self, memberId):
	""" doot de doo """
	return exUserFolder.membershipSources[memberId].manage_addForm

def getAuthSources(self):
	""" Hrm I need a docstring """
	l=[]
	for o in exUserFolder.authSources.keys():
		l.append(
			exUserFolder.authSources[o]
			)
	return l

def getPropSources(self):
	""" Hrm I need a docstring """
	l=[]
	for o in exUserFolder.propSources.keys():
		l.append(
			exUserFolder.propSources[o]
			)
	return l

def getMembershipSources(self):
	""" Hrm I need a docstring """
	l=[]
	for o in exUserFolder.membershipSources.keys():
		l.append(
			exUserFolder.membershipSources[o]
			)
	return l

def MailHostIDs(self):
    """Find SQL database connections in the current folder and above

    This function return a list of ids.
    """
    ids={}
    have_id=ids.has_key
    StringType=type('')

    while self is not None:
        if hasattr(self, 'objectValues'):
            for o in self.objectValues():
                if (hasattr(o,'meta_type') and o.meta_type=='Mail Host'
                    and hasattr(o,'id')):
                    id=o.id
                    if type(id) is not StringType: id=id()
                    if not have_id(id):
                        if hasattr(o,'title_and_id'): o=o.title_and_id()
                        else: o=id
                        ids[id]=id
        if hasattr(self, 'aq_parent'): self=self.aq_parent
        else: self=None

    ids=map(lambda item: (item[1], item[0]), ids.items())
    ids.sort()
    return ids


from types import ListType, IntType, LongType, FloatType, NoneType, DictType, StringType

def getVariableType(self, o):

	if type(o) == ListType:
		return 'List'
	if type(o) == IntType:
		return 'Int'
	if type(o) == LongType:
		return 'Long'
	if type(o) == FloatType:
		return 'Float'
	if type(o) == NoneType:
		return 'None'
	if type(o) == DictType:
		return 'Dict'
	if type(o) == StringType:
		return 'String'
	return 'Unknown or Restricted'

_postUserCreate='''
<dtml-comment>
Replace this method with whatever you want to do
when a user is created, you can use a Python Script,
or External Method, or keep it as a DTML Method if you
want to
</dtml-comment>
'''
