#
# Extensible User Folder
# 
# SMB Authentication Source for exUserFolder
#
# (C) Copyright 2000,2001 The Internet (Aust) Pty Ltd
# ACN: 082 081 472	ABN: 83 082 081 472
# All Rights Reserved
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Author: Andrew Milton <akm@theinternet.com.au>

# Uses pysmb
# Copyright (C) 2001 Michael Teo <michaelteo@bigfoot.com>

import string, Acquisition

from Globals import HTMLFile, MessageDialog, INSTANCE_HOME

from OFS.Folder import Folder

from Products.exUserFolder.exUserFolder import exUserFolder
from Products.exUserFolder.Plugins import PluginRegister

import smb, nmb, socket
from smb import SessionError

def manage_addsmbAuthSource(self, REQUEST):
	""" Add a smb Auth Source """

	host=REQUEST['smbauth_host']
	domain=REQUEST['smbauth_domain']
	winsserver=REQUEST['smbauth_winsserver']

	ob=smbAuthSource(host, domain, winsserver)
	self._setObject('smbAuthSource', ob, None, None, 0)
	self.currentAuthSource=ob	


manage_addsmbAuthSourceForm=HTMLFile('manage_addsmbAuthSourceForm', globals())
manage_editsmbAuthSourceForm=HTMLFile('manage_editsmbAuthSourceForm', globals())


class smbAuthSource(Folder):
	""" """

	meta_type='Authorisation Source'
	id		 ='smbAuthSource'
	title	 ='SMB Authorisation'
	icon ='misc_/exUserFolder/exUserFolderPlugin.gif'

	manage_editForm=manage_editsmbAuthSourceForm
	manage_tabs=Acquisition.Acquired

	_v_smb=None
	_v_netbios=None
	_v_address=None

	def __init__(self, host, domain, winsserver):

		self._setSMBData(host, domain, winsserver)

	def manage_editAuthSource(self, REQUEST):
		'''Handle output of manage_main - change ZSmb instance properties.
		If REQUEST.secret is None, old secret will be used.'''

		host=REQUEST['host']
		domain=REQUEST['domain']
		winsserver=REQUEST['winsserver']

		self._setSMBData(host, domain, winsserver)

		if REQUEST is not None:
			return self.MessageDialog(
				title = 'Edited',
				message = "Properties for %s changed." % self.id,
				action = 'manage_editAuthSourceForm')
		
	#
	# We don't let you delete, create, or edit users
	#
	def deleteUsers(self, userids):
		pass

	def createUser(self, username, password, roles):
		pass


	def updateUser(self, username, password, roles):
		self.currentPropSource.setUserProperty(username=username, key='_roles', value=roles)
	
	def listUserNames(self):
		pass

	def listUsers(self):
		pass

	def listOneUser(self, username):
		roles=[]
		if self.currentPropSource:
			roles=self.currentPropSource.getUserProperty(username=username, key='_roles', default=[])

		username = username.lower()

		return [{'username':username, 'password':'', 'roles':roles},]

	def getUsers(self):
		pass
	
	def authenticate(self, username, password):
		'Authenticate a username/password combination against the Smb server'
		if username == '':
			return 0
		
		return self._authenticate_retry(username, password, 0)

	def _authenticate_retry(self, username, password, retry_depth):
		try:
			self._getSMB().login(username, password, self._domain)
			return 1
		except (SessionError, nmb.NetBIOSError):
			# LRA: we should log nmb.NetBIOSErrors somewhere. It happens if the host
			# can no longer be found.
			# I believe the site manager will want to know when it happens
			# if we just let the exception pass, the site manager may be locked out
			# of the folder.
			return 0
		except socket.error, (errno, strerror):
			if errno == 32:
				self._resetSMB()
				if retry_depth < 1:
					return self._authenticate_retry(username, password, 1)
				else:
					raise
	
	remoteAuthMethod=authenticate

	def host(self): return self._host
	def domain(self): return self._domain
	def port(self): return self._port
	def retries(self): return self._retries
	def timeout(self): return self._timeout
	def winsserver(self): return self._winsserver

	def _getNetBIOS(self):
		if not self._v_netbios:
			self._v_netbios = nmb.NetBIOS()
			# check if we should use WINS
			if self._winsserver:
				self._v_netbios.set_nameserver(self._winsserver)
		return self._v_netbios

	def _getAddress(self):
		if not self._v_address:
			self._v_address = self._getNetBIOS().gethostbyname(self._host)
			if not self._v_address:
				raise nmb.NetBIOSError("smbAuthorization: '%s' is not known to WINS server '%s'" %
						       (self._host, self._winsserver) )
		return self._v_address

	def _getSMB(self):
		if not self._v_smb:
			self._v_smb = smb.SMB(self._host, self._getAddress()[0].get_ip())
		return self._v_smb
		
	def _setSMBData(self, host, domain, winsserver):

		self._host = host
		self._domain = domain
		self._winsserver = winsserver

		# Reset Smb objects so new values take effect. This is
		# why we don't allow direct access to the attributes
		self._resetSMB()

	def _resetSMB(self):
		self._v_netbios = None
		self._v_smb = None
		self._v_address = None

		# if this doesn't raise an Exception, the smb data is valid
		return self._getSMB()


smbAuthReg=PluginRegister('smbAuthSource',
							 'SMB Authentication Source',
							 smbAuthSource, manage_addsmbAuthSourceForm,
							 manage_addsmbAuthSource,
							 manage_editsmbAuthSourceForm)
exUserFolder.authSources['smbAuthSource']=smbAuthReg
